__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"


def main(args, parser, extra, subparser):

    from oras.main import Client

    client = Client(quiet=args.quiet)
    print(client.version())



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import inspect
import logging as _logging
import os
import platform
import sys
import threading
from pathlib import Path
from typing import Optional, Text, TextIO, Union


class ColorizingStreamHandler(_logging.StreamHandler):
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[%dm"
    BOLD_SEQ = "\033[1m"

    colors = {
        "WARNING": YELLOW,
        "INFO": GREEN,
        "DEBUG": BLUE,
        "CRITICAL": RED,
        "ERROR": RED,
    }

    def __init__(
        self,
        nocolor: bool = False,
        stream: Union[Text, Path, TextIO] = sys.stderr,
        use_threads: bool = False,
    ):
        """
        Create a new ColorizingStreamHandler

        :param nocolor: do not use color
        :type nocolor: bool
        :param stream: stream list to this output
        :type stream: bool
        :param use_threads: use threads! lol
        :type use_threads: bool
        """
        super().__init__(stream=stream)
        self._output_lock = threading.Lock()
        self.nocolor = nocolor or not self.can_color_tty()

    def can_color_tty(self) -> bool:
        """
        Determine if the tty supports color
        """
        if "TERM" in os.environ and os.environ["TERM"] == "dumb":
            return False
        return self.is_tty and not platform.system() == "Windows"

    @property
    def is_tty(self) -> bool:
        """
        Determine if we have a tty environment
        """
        isatty = getattr(self.stream, "isatty", None)
        return isatty and isatty()  # type: ignore

    def emit(self, record: _logging.LogRecord):
        """
        Emit a log record

        :param record: the record to emit
        :type record: logging.LogRecord
        """
        with self._output_lock:
            try:
                self.format(record)  # add the message to the record
                self.stream.write(self.decorate(record))
                self.stream.write(getattr(self, "terminator", "\n"))
                self.flush()
            except BrokenPipeError as e:
                raise e
            except (KeyboardInterrupt, SystemExit):
                # ignore any exceptions in these cases as any relevant messages have been printed before
                pass
            except Exception:
                self.handleError(record)

    def decorate(self, record) -> str:
        """
        Decorate a log record

        :param record: the record to emit
        """
        message = record.message
        message = [message]
        if not self.nocolor and record.levelname in self.colors:
            message.insert(0, self.COLOR_SEQ % (30 + self.colors[record.levelname]))
            message.append(self.RESET_SEQ)
        return "".join(message)


class Logger:
    def __init__(self):
        """
        Create a new logger
        """
        self.logger = _logging.getLogger(__name__)
        self.log_handler = [self.text_handler]
        self.stream_handler = None
        self.printshellcmds = False
        self.quiet = False
        self.logfile = None
        self.last_msg_was_job_info = False
        self.logfile_handler = None

    def cleanup(self):
        """
        Close open files, etc. for the logger
        """
        if self.logfile_handler is not None:
            self.logger.removeHandler(self.logfile_handler)
            self.logfile_handler.close()
        self.log_handler = [self.text_handler]

    def handler(self, msg: dict):
        """
        Handle a log message.

        :param msg: the message to handle
        :type msg: dict
        """
        for handler in self.log_handler:
            handler(msg)

    def set_stream_handler(self, stream_handler: _logging.Handler):
        """
        Set a stream handler.

        :param stream_handler : the stream handler
        :type stream_handler: logging.Handler
        """
        if self.stream_handler is not None:
            self.logger.removeHandler(self.stream_handler)
        self.stream_handler = stream_handler
        self.logger.addHandler(stream_handler)

    def set_level(self, level: int):
        """
        Set the logging level.

        :param level: the logging level to set
        :type level: int
        """
        self.logger.setLevel(level)

    def location(self, msg: str):
        """
        Debug level message with location info.

        :param msg: the logging message
        :type msg: dict
        """
        callerframerecord = inspect.stack()[1]
        frame = callerframerecord[0]
        info = inspect.getframeinfo(frame)
        self.debug(
            "{}: {info.filename}, {info.function}, {info.lineno}".format(msg, info=info)
        )

    def info(self, msg: str):
        """
        Info level message

        :param msg: the informational message
        :type msg: str
        """
        self.handler({"level": "info", "msg": msg})

    def warning(self, msg: str):
        """
        Warning level message

        :param msg: the warning message
        :type msg: str
        """
        self.handler({"level": "warning", "msg": msg})

    def debug(self, msg: str):
        """
        Debug level message

        :param msg: the debug message
        :type msg: str
        """
        self.handler({"level": "debug", "msg": msg})

    def error(self, msg: str):
        """
        Error level message

        :param msg: the error message
        :type msg: str
        """
        self.handler({"level": "error", "msg": msg})

    def exit(self, msg: str, return_code: int = 1):
        """
        Error level message and exit with error code

        :param msg: the exiting (error) message
        :type msg: str
        :param return_code: return code to exit on
        :type return_code: int
        """
        self.handler({"level": "error", "msg": msg})
        sys.exit(return_code)

    def progress(self, done: int = None, total: int = None):
        """
        Show piece of a progress bar

        :param done: count of total that is complete
        :type done: int
        :param total: count of total
        :type total: int
        """
        self.handler({"level": "progress", "done": done, "total": total})

    def shellcmd(self, msg: Optional[str]):
        """
        Shellcmd message

        :param msg: the message
        :type msg: str
        """
        if msg is not None:
            self.handler({"level": "shellcmd", "msg": msg})

    def text_handler(self, msg: dict):
        """
        The default log handler that prints to the console.

        :param msg: the log message dict
        :type msg: dict
        """
        level = msg["level"]
        if level == "info" and not self.quiet:
            self.logger.info(msg["msg"])
        if level == "warning":
            self.logger.warning(msg["msg"])
        elif level == "error":
            self.logger.error(msg["msg"])
        elif level == "debug":
            self.logger.debug(msg["msg"])
        elif level == "progress" and not self.quiet:
            done = msg["done"]
            total = msg["total"]
            p = done / total
            percent_fmt = ("{:.2%}" if p < 0.01 else "{:.0%}").format(p)
            self.logger.info(
                "{} of {} steps ({}) done".format(done, total, percent_fmt)
            )
        elif level == "shellcmd":
            if self.printshellcmds:
                self.logger.warning(msg["msg"])


logger = Logger()


def setup_logger(
    quiet: bool = False,
    printshellcmds: bool = False,
    nocolor: bool = False,
    stdout: bool = False,
    debug: bool = False,
    use_threads: bool = False,
):
    """
    Setup the logger. This should be called from an init or client.

    :param quiet: set logging level to quiet
    :type quiet: bool
    :param printshellcmds: a special level to print shell commands
    :type printshellcmds: bool
    :param nocolor: do not use color
    :type nocolor: bool
    :param stdout: print to standard output for the logger
    :type stdout: bool
    :param debug: debug level logging
    :type debug: bool
    :param use_threads: use threads!
    :type use_threads: bool
    """
    # console output only if no custom logger was specified
    stream_handler = ColorizingStreamHandler(
        nocolor=nocolor,
        stream=sys.stdout if stdout else sys.stderr,
        use_threads=use_threads,
    )
    level = _logging.INFO
    if debug:
        level = _logging.DEBUG

    logger.set_stream_handler(stream_handler)
    logger.set_level(level)
    logger.quiet = quiet
    logger.printshellcmds = printshellcmds



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import oras.client


def main(args, parser, extra, subparser):
    """
    Main is a light wrapper around the logout command.
    """
    client = oras.client.OrasClient()
    client.logout(args.hostname)



#!/usr/bin/env python

__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import argparse
import os
import sys

import oras
import oras.cli.help as help
from oras.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="OCI Registry as Storage Python client",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="Show the oras version information.",
        default=False,
        action="store_true",
    )

    description = "actions for ORAS Python"
    subparsers = parser.add_subparsers(
        help="oras actions",
        title="actions",
        description=description,
        dest="command",
        required=True,
    )

    # print version and exit
    subparsers.add_parser("version", description="show software version")

    # Login
    login = subparsers.add_parser(
        "login",
        description=help.login_help,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    logout = subparsers.add_parser("logout", description="logout from a registry")

    login.add_argument(
        "--password-stdin",
        dest="password_stdin",
        help="read password or identity token from stdin",
        default=False,
        action="store_true",
    )

    # Login and logout share config and hostname arguments
    for command in login, logout:
        command.add_argument("hostname", help="hostname")

    pull = subparsers.add_parser("pull", description="pull a container")
    pull.add_argument(
        "--allowed-media-type", help="add an allowed media type.", action="append"
    )
    pull.add_argument(
        "--allow-all-media-types",
        help="allow all media types",
        default=False,
        action="store_true",
    )
    pull.add_argument(
        "--keep-old-files",
        help="do not overwrite existing files.",
        default=False,
        action="store_true",
    )
    # todo we haven't added path traversal, or cacheRoot to pull
    pull.add_argument("--output", help="output directory.", default=os.getcwd())
    pull.add_argument("--manifest-config-ref", help="manifest config reference")

    push = subparsers.add_parser("push", description=help.push_help)
    push.add_argument("--manifest-annotations", help="manifest annotation file")
    push.add_argument("--manifest-config", help="manifest config file")
    push.add_argument(
        "--disable-path-validation",
        help="skip path validation",
        default=False,
        action="store_true",
    )
    for command in push, pull:
        command.add_argument("target", help="target")
    push.add_argument("filerefs", help="file references", nargs="+")

    # Debug is added on the level of the command
    for command in login, logout, push, pull:
        command.add_argument(
            "--debug",
            dest="debug",
            help="debug mode",
            default=False,
            action="store_true",
        )

    for command in login, logout, push, pull:
        command.add_argument(
            "-c",
            "--config",
            dest="config",
            help="auth config path",
            action="append",
        )

    # login and push/pull share username/password, and insecure
    for command in login, push, pull:
        command.add_argument(
            "-u", "--username", dest="username", help="registry username"
        )
        command.add_argument(
            "-p",
            "--password",
            dest="password",
            help="registry password or identity token",
        )
        command.add_argument(
            "-k",
            "--insecure",
            dest="insecure",
            help="allow connections to SSL registry without certs",
            default=False,
            action="store_true",
        )

    shell = subparsers.add_parser("shell", description="create an interactive shell")
    return parser


def run():
    """
    Entrypoint to ORAS Python
    """
    parser = get_parser()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    def help(return_code=0):
        version = oras.__version__

        print("\nOCI Registry as Storage (oras) Python Client v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # retrieve subparser (with help) from parser
    helper = None
    subparsers_actions = [
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    ]
    for subparsers_action in subparsers_actions:
        for choice, subparser in subparsers_action.choices.items():
            if choice == args.command:
                helper = subparser

    if args.debug:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    setup_logger(quiet=args.quiet, debug=args.debug)

    # Direct to the right parser
    if args.command == "version" or args.version:
        from .version import main
    elif args.command == "login":
        from .login import main
    elif args.command == "logout":
        from .logout import main
    elif args.command == "pull":
        from .pull import main
    elif args.command == "push":
        from .push import main
    elif args.command == "shell":
        from .shell import main

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args=args, parser=parser, extra=extra, subparser=helper)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)


if __name__ == "__main__":
    run()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import copy
import os
from typing import List, Optional, Tuple, Union

import jsonschema
import requests

import oras.auth
import oras.container
import oras.oci
import oras.schemas
import oras.utils
from oras.decorator import ensure_container
from oras.logger import logger


class Registry:
    """
    Direct interactions with an OCI registry.

    This could also be called a "provider" when we add in the "copy" logic
    and the registry isn't necessarily the "remote" endpoint.
    """

    def __init__(self, hostname: Optional[str] = None, insecure: bool = False):
        """
        Create a new registry provider.

        :param hostname: the registry hostname (optional)
        :type hostname: str
        :param insecure: use http instead of https
        :type insecure: bool
        """
        self.hostname: Optional[str] = hostname
        self.headers: dict = {}
        self.session: requests.Session = requests.Session()
        self.prefix: str = "http" if insecure else "https"
        self.token: Optional[str] = None
        self._auths: dict = {}

    def logout(self, hostname: str):
        """
        If auths are loaded, remove a hostname.

        :param hostname: the registry hostname to remove
        :type hostname: str
        """
        if not self._auths:
            logger.info(f"You are not logged in to {hostname}")
            return

        for host in oras.utils.iter_localhosts(hostname):
            if host in self._auths:
                del self._auths[host]
                logger.info(f"You have successfully logged out of {hostname}")
                return
        logger.info(f"You are not logged in to {hostname}")

    @ensure_container
    def load_configs(
        self,
        container: Union[str, oras.container.Container],
        configs: Optional[list] = None,
    ):
        """
        Load configs to discover credentials for a specific container.

        This is typically just called once. We always add the default Docker
        config to the set.s

        :param container: the parsed container URI with components
        :type container: oras.container.Container
        :param configs: list of configs to read (optional)
        :type configs: list
        """
        if not self._auths:
            self._auths = oras.auth.load_configs(configs)
        for registry in oras.utils.iter_localhosts(container.registry):  # type: ignore
            if self._load_auth(registry):
                return

    def _load_auth(self, hostname: str) -> bool:
        """
        Look for and load a named authentication token.

        :param hostname: the registry hostname to look for
        :type hostname: str
        """
        if hostname in self._auths:
            self.token = self._auths[hostname]["auth"]
            return True
        return False

    def set_basic_auth(self, username: str, password: str):
        """
        Set basic authentication.

        :param username: the user account name
        :type username: str
        :param password: the user account password
        :type password: str
        """
        basic_auth = oras.auth.get_basic_auth(username, password)
        self.set_header("Authorization", "Basic %s" % basic_auth)

    def set_header(self, name: str, value: str):
        """
        Courtesy function to set a header

        :param name: header name to set
        :type name: str
        :param value: header value to set
        :type value: str
        """
        self.headers.update({name: value})

    def _validate_path(self, path: str) -> bool:
        """
        Ensure a blob path is in the present working directory or below.

        :param path: the path to validate
        :type path: str
        """
        return os.getcwd() in os.path.abspath(path)

    def _parse_manifest_ref(self, ref: str) -> Union[Tuple[str, str], List[str]]:
        """
        Parse an optional manifest config, e.g:

        Examples
        --------
        <path>:<content-type>
        path/to/config:application/vnd.oci.image.config.v1+json
        /dev/null:application/vnd.oci.image.config.v1+json

        :param ref: the manifest reference to parse (examples above)
        :type ref: str
        """
        if ":" not in ref:
            return ref, oras.defaults.unknown_config_media_type
        return ref.split(":", 1)

    def _upload_blob(
        self, blob: str, container: Union[str, oras.container.Container], layer: dict
    ) -> requests.Response:
        """
        Prepare and upload a blob.

        Sizes > 1024 are uploaded via a chunked approach (post, patch+, put)
        and <= 1024 is a single post then put.

        :param blob: path to blob to upload
        :type blob: str
        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param layer: dict from oras.oci.NewLayer
        :type layer: dict
        """
        blob = os.path.abspath(blob)
        container = self.get_container(container)
        size = layer["size"]

        # Chunked for large, otherwise POST and PUT
        if size < 1024:
            return self._put_upload(blob, container, layer)
        return self._chunked_upload(blob, container, layer)

    @ensure_container
    def get_tags(
        self, container: Union[str, oras.container.Container], N: int = 10_000
    ) -> List[str]:
        """
        Retrieve tags for a package.

        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param N: number of tags
        :type N: int
        """
        tags_url = f"{self.prefix}://{container.tags_url(N)}"  # type: ignore
        return self.do_request(tags_url, "GET", headers=self.headers)

    @ensure_container
    def get_blob(
        self,
        container: Union[str, oras.container.Container],
        digest: str,
        stream: bool = False,
    ) -> requests.Response:
        """
        Retrieve a blob for a package.

        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param digest: sha256 digest of the blob to retrieve
        :type digest: str
        :param stream: stream the response (or not)
        :type stream: bool
        """
        blob_url = f"{self.prefix}://{container.get_blob_url(digest)}"  # type: ignore
        return self.do_request(blob_url, "GET", headers=self.headers, stream=stream)

    def get_container(
        self, name: Union[str, oras.container.Container]
    ) -> oras.container.Container:
        """
        Courtesy function to get a container from a URI.

        :param name: unique resource identifier to parse
        :type name: oras.container.Container or str
        """
        if isinstance(name, oras.container.Container):
            return name
        return oras.container.Container(name, registry=self.hostname)

    @ensure_container
    def _download_blob(
        self, container: Union[str, oras.container.Container], digest: str, outfile: str
    ) -> str:
        """
        Stream download a blob into an output file.

        This function is a wrapper around get_blob.

        :param container:  parsed container URI
        :type container: oras.container.Container or str
        """
        with self.get_blob(container, digest, stream=True) as r:
            r.raise_for_status()
            with open(outfile, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        return outfile

    def _put_upload(
        self, blob: str, container: oras.container.Container, layer: dict
    ) -> requests.Response:
        """
        Upload to a registry via put.

        :param blob: path to blob to upload
        :type blob: str
        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param layer: dict from oras.oci.NewLayer
        :type layer: dict
        """
        # Start an upload session
        headers = {"Content-Type": "application/octet-stream"}
        upload_url = f"{self.prefix}://{container.upload_blob_url()}"
        r = self.do_request(upload_url, "POST", headers=headers)

        # Location should be in the header
        session_url = r.headers.get("location")

        # PUT to upload blob url
        headers = {
            "Content-Length": str(layer["size"]),
            "Content-Type": "application/octet-stream",
        }

        # headers.update(auth_headers)
        digest = layer["digest"]
        blob_url = f"{session_url}&digest={digest}"
        with open(blob, "rb") as fd:
            response = self.do_request(
                blob_url, method="PUT", data=fd.read(), headers=headers
            )
        return response

    def _chunked_upload(
        self, blob: str, container: oras.container.Container, layer: dict
    ) -> requests.Response:
        """
        Upload via a chunked upload.

        :param blob: path to blob to upload
        :type blob: str
        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param layer: dict from oras.oci.NewLayer
        :type layer: dict
        """
        # Start an upload session
        headers = {"Content-Type": "application/octet-stream", "Content-Length": 0}
        upload_url = f"{self.prefix}://{container.upload_blob_url()}"
        r = self.do_request(upload_url, "POST", headers=headers)

        # Location should be in the header
        session_url = r.headers.get("location")

        # Read the blob in chunks, for each do a patch
        start = 0
        with open(blob, "rb") as fd:
            for chunk in oras.utils.read_in_chunks(fd):
                if not chunk:
                    break

                end = start + len(chunk) - 1
                content_range = "%s-%s" % (start, end)
                headers = {
                    "Content-Range": content_range,
                    "Content-Length": str(len(chunk)),
                    "Content-Type": "application/octet-stream",
                }
                # headers.update(auth_headers)
                start = end + 1
                self._check_200_response(
                    self.do_request(session_url, "PATCH", data=chunk, headers=headers)
                )

        # Finally, issue a PUT request to close blob
        session_url = "%s&digest=%s" % (session_url, layer["digest"])
        return self.do_request(session_url, "PUT")

    def _check_200_response(self, response: requests.Response):
        """
        Helper function to ensure some flavor of 200

        :param response: request response to inspect
        :type response: requests.Response
        """
        if response.status_code not in [200, 201, 202]:
            self._parse_response_errors(response)
            logger.exit(f"Issue with {response.request.url}:\n{response.reason}")

    def _parse_response_errors(self, response: requests.Response):
        """
        Given a failed request, look for OCI formatted error messages.

        :param response: request response to inspect
        :type response: requests.Response
        """
        try:
            msg = response.json()
            for error in msg.get("errors", []):
                if isinstance(error, dict) and "message" in error:
                    logger.error(error["message"])
        except:
            pass

    def _upload_manifest(
        self, manifest: dict, container: oras.container.Container
    ) -> requests.Response:
        """
        Read a manifest file and upload it.

        :param manifest: manifest to upload
        :type manifest: dict
        :param container:  parsed container URI
        :type container: oras.container.Container or str
        """
        jsonschema.validate(manifest, schema=oras.schemas.manifest)
        headers = {
            "Content-Type": oras.defaults.default_manifest_media_type,
            "Content-Length": str(len(manifest)),
        }
        put_url = f"{self.prefix}://{container.put_manifest_url()}"
        return self.do_request(put_url, "PUT", headers=headers, json=manifest)

    def push(self, *args, **kwargs) -> requests.Response:
        """
        Push a set of files to a target

        :param config_path: path to a config file
        :type config_path: str
        :param disable_path_validation: ensure paths are relative to the running directory.
        :type disable_path_validation: bool
        :param files: list of files to push
        :type files: list
        :param insecure: allow registry to use http
        :type insecure: bool
        :param manifest_config: content type
        :type manifest_config: str
        :param manifest_annotations: manifest annotations file
        :type manifest_annotations: str
        :param username: username for basic auth
        :type username: str
        :param password: password for basic auth
        :type password: str
        :param target: target location to push to
        :type target: str
        """
        container = self.get_container(kwargs["target"])
        self.load_configs(container, configs=kwargs.get("config_path"))

        # Hold state of request for http/https
        validate_path = not kwargs.get("disable_path_validation", False)

        # Prepare a new manifest
        manifest = oras.oci.NewManifest()

        # A lookup of annotations we can add
        annotset = oras.oci.Annotations(kwargs.get("manifest_annotations"))

        # Upload files as blobs
        for blob in kwargs.get("files", []):

            # Must exist
            if not os.path.exists(blob):
                logger.exit(f"{blob} does not exist.")

            # Path validation means blob must be relative to PWD.
            if validate_path:
                if not self._validate_path(blob):
                    logger.exit(
                        f"Blob {blob} is not in the present working directory context."
                    )

            # Save directory or blob name before compressing
            blob_name = os.path.basename(blob)

            # If it's a directory, we need to compress
            cleanup_blob = False
            if os.path.isdir(blob):
                blob = oras.utils.make_targz(blob)
                cleanup_blob = True

            # Create a new layer from the blob
            layer = oras.oci.NewLayer(blob, is_dir=cleanup_blob)
            annotations = annotset.get_annotations(blob)
            layer["annotations"] = {oras.defaults.annotation_title: blob_name}
            if annotations:
                layer["annotations"].update(annotations)

            # update the manifest with the new layer
            manifest["layers"].append(layer)

            # Upload the blob layer
            response = self._upload_blob(blob, container, layer)
            self._check_200_response(response)

            # Do we need to cleanup a temporary targz?
            if cleanup_blob and os.path.exists(blob):
                os.remove(blob)

        # Add annotations to the manifest, if provided
        manifest_annots = annotset.get_annotations("$manifest")
        if manifest_annots:
            manifest["annotations"] = manifest_annots

        # Prepare the manifest config (temporary or one provided)
        manifest_config = kwargs.get("manifest_config")
        config_annots = annotset.get_annotations("$config")
        if manifest_config:
            ref, media_type = self._parse_manifest_ref(manifest_config)
            conf, config_file = oras.oci.ManifestConfig(ref, media_type)
        else:
            conf, config_file = oras.oci.ManifestConfig()

        # Config annotations?
        if config_annots:
            conf["annotations"] = config_annots

        # Config is just another layer blob!
        response = self._upload_blob(config_file, container, conf)
        self._check_200_response(response)

        # Final upload of the manifest
        manifest["config"] = conf
        self._check_200_response(self._upload_manifest(manifest, container))
        print(f"Successfully pushed {container}")
        return response

    def pull(self, *args, **kwargs) -> List[str]:
        """
        Pull an artifact from a target

        :param config_path: path to a config file
        :type config_path: str
        :param allowed_media_type: list of allowed media types
        :type allowed_media_type: list or None
        :param overwrite: if output file exists, overwrite
        :type overwrite: bool
        :param manifest_config_ref: sav manifest config to this file
        :type manifest_config_ref: str
        :param outdir: output directory path
        :type outdir: str
        :param username: username for basic auth
        :type username: str
        :param password: password for basic auth
        :type password: str
        :param target: target location to pull from
        :type target: str
        """
        allowed_media_type = kwargs.get("allowed_media_type")
        container = self.get_container(kwargs["target"])
        self.load_configs(container, configs=kwargs.get("config_path"))
        manifest = self.get_manifest(container, allowed_media_type)
        outdir = kwargs.get("outdir") or oras.utils.get_tmpdir()
        overwrite = kwargs.get("overwrite", True)

        files = []
        for layer in manifest.get("layers", []):
            filename = layer.get("annotations", {}).get(oras.defaults.annotation_title)

            # If we don't have a filename, default to digest. Hopefully does not happen
            if not filename:
                filename = layer["digest"]
            outfile = os.path.join(outdir, filename)
            if not overwrite and os.path.exists(outfile):
                logger.warning(
                    f"{outfile} already exists and --keep-old-files set, will not overwrite."
                )
                continue

            # A directory will need to be uncompressed and moved
            if layer["mediaType"] == oras.defaults.default_blob_dir_media_type:
                targz = oras.utils.get_tmpfile(suffix=".tar.gz")
                self._download_blob(container, layer["digest"], targz)

                # The artifact will be extracted to the correct name
                oras.utils.extract_targz(targz, os.path.dirname(outfile))

            # Anything else just extracted directly
            else:
                self._download_blob(container, layer["digest"], outfile)
            logger.info(f"Successfully pulled {outfile}.")
            files.append(outfile)
        return files

    @ensure_container
    def get_manifest(
        self,
        container: Union[str, oras.container.Container],
        allowed_media_type: list = None,
    ) -> dict:
        """
        Retrieve a manifest for a package.

        :param container:  parsed container URI
        :type container: oras.container.Container or str
        :param allowed_media_type: one or more allowed media types
        :type allowed_media_type: str
        """
        if not allowed_media_type:
            allowed_media_type = [oras.defaults.default_manifest_media_type]
        headers = {"Accept": ";".join(allowed_media_type)}
        url = f"{self.prefix}://{container.get_manifest_url()}"  # type: ignore
        response = self.do_request(url, "GET", headers=headers)
        self._check_200_response(response)
        manifest = response.json()
        jsonschema.validate(manifest, schema=oras.schemas.manifest)
        return manifest

    def do_request(
        self,
        url: str,
        method: str = "GET",
        data: Union[dict, bytes] = None,
        headers: dict = None,
        json: dict = None,
        stream: bool = False,
    ):
        """
        Do a request. This is a wrapper around requests to handle retry auth.

        :param url: the URL to issue the request to
        :type url: str
        :param method: the method to use (GET, DELETE, POST, PUT, PATCH)
        :type method: str
        :param data: data for requests
        :type data: dict or bytes
        :param headers: headers for the request
        :type headers: dict
        :param json: json data for requests
        :type json: dict
        :param stream: stream the responses
        :type stream: bool
        """
        headers = headers or {}

        # Make the request and return to calling function, unless requires auth
        response = self.session.request(
            method, url, data=data, json=json, headers=headers, stream=stream
        )

        # A 401 response is a request for authentication
        if response.status_code != 401:
            return response

        # Otherwise, authenticate the request and retry
        if self.authenticate_request(response):
            headers.update(self.headers)
            return self.session.request(
                method, url, data=data, json=json, headers=headers, stream=stream
            )
        return response

    def authenticate_request(self, originalResponse: requests.Response) -> bool:
        """
        Authenticate Request
        Given a response, look for a Www-Authenticate header to parse.

        We return True/False to indicate if the request should be retried.

        :param originalResponse: original response to get the Www-Authenticate header
        :type originalResponse: requests.Response
        """
        authHeaderRaw = originalResponse.headers.get("Www-Authenticate")
        if not authHeaderRaw:
            return False

        # If we have a token, set auth header (base64 encoded user/pass)
        if self.token:
            self.set_header("Authorization", "Basic %s" % self.token)

        headers = copy.deepcopy(self.headers)
        if "Authorization" not in headers:
            logger.error(
                "This endpoint requires a token. Please set "
                "oras.provider.Registry.set_basic_auth(username, password) "
                "first or use oras-py login to do the same."
            )
            return False

        # Prepare request to retry
        h = oras.auth.parse_auth_header(authHeaderRaw)
        if h.service:
            headers.update(
                {
                    "Service": h.service,
                    "Accept": "application/json",
                    "User-Agent": "oras-py",
                }
            )

        # Currently we don't set a scope (it defaults to build)
        if not h.realm.startswith("http"):  # type: ignore
            h.realm = f"{self.prefix}://{h.realm}"
        authResponse = self.session.get(h.realm, headers=headers)  # type: ignore
        if authResponse.status_code != 200:
            return False

        # Request the token
        info = authResponse.json()
        token = info.get("token")
        if not token:
            token = info.get("access_token")

        # Set the token to the original request and retry
        self.headers.update({"Authorization": "Bearer %s" % token})
        return True



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"


import re
from typing import Optional

import oras.defaults
from oras.logger import logger

docker_regex = re.compile(
    "(?:(?P<registry>[^/@]+[.:][^/@]*)/)?"
    "(?P<namespace>(?:[^:@/]+/)+)?"
    "(?P<repository>[^:@/]+)"
    "(?::(?P<tag>[^:@]+))?"
    "(?:@(?P<digest>.+))?"
    "$"
)


class Container:
    def __init__(self, name: str, registry: Optional[str] = None):
        """
        Parse a container name and easily get urls for registry interactions.

        :param name: the full name of the container to parse (with any components)
        :type name: str
        :param registry: a custom registry name, if not provided with URI
        :type registry: str
        """
        self.registry = registry or oras.defaults.registry.index_name

        # Registry is the name takes precendence
        self.parse(name)

    def get_blob_url(self, digest: str) -> str:
        """
        Get the URL to download a blob

        :param digest: the digest to download
        :type digest: str
        """
        return f"{self.registry}/v2/{self.namespace}/{self.repository}/blobs/{digest}"

    def upload_blob_url(self) -> str:
        return f"{self.registry}/v2/{self.namespace}/{self.repository}/blobs/uploads/"

    def tags_url(self, N=10_000) -> str:
        return f"{self.registry}/v2/{self.namespace}/{self.repository}/tags/list?n={N}"

    def put_manifest_url(self) -> str:
        return f"{self.registry}/v2/{self.namespace}/{self.repository}/manifests/{self.tag}"

    def get_manifest_url(self) -> str:
        return f"{self.registry}/v2/{self.namespace}/{self.repository}/manifests/{self.tag}"

    def __str__(self) -> str:
        return self.uri

    @property
    def uri(self) -> str:
        """
        Assemble the complete unique resource identifier
        """
        uri = f"{self.namespace}/{self.repository}"
        if self.registry:
            uri = f"{self.registry}/{uri}"

        # Digest takes preference because more specific
        if self.digest:
            uri = f"{uri}@{self.digest}"
        elif self.tag:
            uri = f"{uri}:{self.tag}"
        return uri

    def parse(self, name: str):
        """
        Parse the container name into registry, repository, and tag.

        :param name: the full name of the container to parse (with any components)
        :type name: str
        """
        match = re.search(docker_regex, name)
        if not match:
            logger.exit(
                f"{name} does not match a recognized registry unique resource identifier. Try <registry>/<namespace>/<repository>:<tag|digest>"
            )
        items = match.groupdict()  # type: ignore
        self.repository = items["repository"]
        self.registry = items["registry"] or self.registry
        self.namespace = items["namespace"]
        self.tag = items["tag"] or oras.defaults.default_tag
        self.digest = items["digest"]

        # Repository and namespace are required
        if not self.repository or not self.namespace:
            logger.exit(
                "You are minimally required to include a <namespace>/<repository>"
            )
        self.namespace = self.namespace.strip("/")



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

__version__ = "0.0.1"
AUTHOR = "Vanessa Sochat"
EMAIL = "vsoch@users.noreply.github.com"
NAME = "oras"
PACKAGE_URL = "https://github.com/oras-project/oras-python"
KEYWORDS = "oci, registry, storage"
DESCRIPTION = "OCI Registry as Storage Python client"
LICENSE = "LICENSE"

################################################################################
# Global requirements

INSTALL_REQUIRES = (
    ("jsonschema", {"min_version": None}),
    ("requests", {"min_version": None}),
    ("docker", {"exact_version": "5.0.1"}),
)

TESTS_REQUIRES = (
    ("pytest", {"min_version": "4.6.2"}),
    ("mypy", {"min_version": None}),
    ("pyflakes", {"min_version": None}),
    ("black", {"min_version": None}),
    ("types-requests", {"min_version": None}),
    ("isort", {"min_version": None}),
)
INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

login_help = """
Log in to a remote registry

Example - Login with username and password from command line:
  oras-py login -u username -p password localhost:5000
Example - Login with username and password from stdin:
  oras-py login -u username --password-stdin localhost:5000
Example - Login with identity token from command line:
  oras-py login -p token localhost:5000
Example - Login with identity token from stdin:
  oras-py login --password-stdin localhost:5000
Example - Login with username and password by prompt:
  oras-py login localhost:5000
Example - Login with insecure registry from command line:
  oras-py login --insecure localhost:5000
"""

copy_help = """Copy artifacts from one location to another",

Example - Copy artifacts from local files to local files:
  oras copy foo/bar:v1 --from files --to files:path/to/save file1 file2 ... filen
Example - Copy artifacts from registry to local files:
  oras copy foo/bar:v1 --from registry --to files:path/to/save
Example - Copy artifacts from registry to oci:
  oras copy foo/bar:v1 --from registry --to oci:path/to/oci
Example - Copy artifacts from local files to registry:
  oras copy foo/bar:v1 --from files --to registry file1 file2 ... filen
When the source (--from) is "files", the config by default will be "{}" and of media type
application/vnd.unknown.config.v1+json. You can override it by setting the path, for example:
  oras copy foo/bar:v1 --from files --manifest-config path/to/config:application/vnd.oci.image.config.v1+json --to files:path/to/save file1 file2 ... fileN"""


logout_help = """
Log out from a remote registry

Example - Logout:
  oras-py logout localhost:5000
"""

push_help = """
Push files to remote registry

Example - Push file "hi.txt" with the "application/vnd.oci.image.layer.v1.tar" media type (default):
  oras-py push localhost:5000/hello:latest hi.txt
Example - Push file "hi.txt" with the custom "application/vnd.me.hi" media type:
  oras-py push localhost:5000/hello:latest hi.txt:application/vnd.me.hi
Example - Push multiple files with different media types:
  oras-py push localhost:5000/hello:latest hi.txt:application/vnd.me.hi bye.txt:application/vnd.me.bye
Example - Push file "hi.txt" with the custom manifest config "config.json" of the custom "application/vnd.me.config" media type:
  oras-py push --manifest-config config.json:application/vnd.me.config localhost:5000/hello:latest hi.txt
Example - Push file to the insecure registry:
  oras-py push localhost:5000/hello:latest hi.txt --insecure
Example - Push file to the HTTP registry:
  oras-py push localhost:5000/hello:latest hi.txt --plain-http
"""



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import oras.client


def main(args, parser, extra, subparser):
    """
    Main is a light wrapper around the login command.
    """
    client = oras.client.OrasClient()
    client.login(
        password=args.password,
        username=args.username,
        config_path=args.config,
        hostname=args.hostname,
        insecure=args.insecure,
        password_stdin=args.password_stdin,
    )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import os
import shutil
import sys

import pytest

import oras.client

here = os.path.abspath(os.path.dirname(__file__))

registry_host = os.environ.get("ORAS_HOST")
registry_port = os.environ.get("ORAS_PORT")


def setup_module(module):
    """
    Ensure the registry port and host is in the environment.
    """
    if not registry_host or not registry_port:
        sys.exit(
            "You must export ORAS_HOST and ORAS_PORT for a running registry before running tests."
        )


registry = f"{registry_host}:{registry_port}"
target = f"{registry}/dinosaur/artifact:v1"
target_dir = f"{registry}/dinosaur/directory:v1"


def test_basic_oras(tmp_path):
    """
    Basic tests for oras (without authentication)
    """
    client = oras.client.OrasClient(insecure=True)
    assert "Python version" in client.version()


def test_basic_push_pull(tmp_path):
    """
    Basic tests for oras (without authentication)
    """
    client = oras.client.OrasClient(insecure=True)
    artifact = os.path.join(here, "artifact.txt")

    assert os.path.exists(artifact)
    res = client.push(files=[artifact], target=target)
    assert res.status_code == 201

    # Test getting tags
    tags = client.get_tags(target)
    for key in ["name", "tags"]:
        assert key in tags
    assert "v1" in tags["tags"]

    # Test pulling elsewhere
    files = client.pull(target=target, outdir=tmp_path)
    assert len(files) == 1
    assert os.path.basename(files[0]) == "artifact.txt"
    assert str(tmp_path) in files[0]
    assert os.path.exists(files[0])

    # Move artifact outside of context (should not work)
    moved_artifact = tmp_path / os.path.basename(artifact)
    shutil.copyfile(artifact, moved_artifact)
    with pytest.raises(SystemExit):
        client.push(files=[moved_artifact], target=target)

    # This should work because we aren't checking paths
    res = client.push(files=[artifact], target=target, disable_path_validation=True)
    assert res.status_code == 201


def test_directory_push_pull(tmp_path):
    """
    Test push and pull for directory
    """
    client = oras.client.OrasClient(insecure=True)

    # Test upload of a directory
    upload_dir = os.path.join(here, "upload_data")
    res = client.push(files=[upload_dir], target=target_dir)
    assert res.status_code == 201
    files = client.pull(target=target_dir, outdir=tmp_path)

    assert len(files) == 1
    assert os.path.basename(files[0]) == "upload_data"
    assert str(tmp_path) in files[0]
    assert os.path.exists(files[0])
    assert "artifact.txt" in os.listdir(files[0])



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"


import sys
from typing import List, Optional, Union

import oras.container
import oras.main as main
import oras.provider
import oras.version


class OrasClient:
    """
    Create an OCI Registry as Storage (ORAS) Client.

    This is intended for controlled interactions. The user of oras-py can use
    this client, the terminal command line wrappers, or the functions in main
    in isolation as an internal Python API. The user can provide a custom
    registry as a parameter, if desired. If not provided we default to standard
    oras.
    """

    def __init__(
        self,
        hostname: Optional[str] = None,
        registry: Optional[oras.provider.Registry] = None,
        insecure: bool = False,
    ):
        """
        Create an ORAS client.

        The hostname is the remote registry to ping.

        :param hostname: the hostname of the registry to ping
        :type hostname: str
        :param registry: if provided, use this custom provider instead of default
        :type registry: oras.provider.Registry or None
        :param insecure: use http instead of https
        :type insecure: bool
        """
        self.remote = registry or oras.provider.Registry(hostname, insecure)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return "[oras-client]"

    def set_basic_auth(self, username: str, password: str):
        """
        Add basic authentication to the request.

        :param username: the user account name
        :type username: str
        :param password: the user account password
        :type password: str
        """
        self.remote.set_basic_auth(username, password)

    def version(self, return_items: bool = False) -> Union[dict, str]:
        """
        Get the version of the client.

        :param return_items : return the dict of version info instead of string
        :type return_items: bool
        """
        version = oras.version.__version__

        python_version = "%s.%s.%s" % (
            sys.version_info.major,
            sys.version_info.minor,
            sys.version_info.micro,
        )
        versions = {"Version": version, "Python version": python_version}

        # If the user wants the dictionary of items returned
        if return_items:
            return versions

        # Otherwise return a string that can be printed
        return "\n".join(["%s: %s" % (k, v) for k, v in versions.items()])

    def get_tags(self, name: str, N: int = 10_000) -> List[str]:
        """
        Retrieve tags for a package.

        :param name: container URI to parse
        :type name: str
        :param N: number of tags
        :type N: int
        """
        return self.remote.get_tags(name, N=N).json()

    def push(self, *args, **kwargs):
        """
        Push a container to the remote.
        """
        return self.remote.push(*args, **kwargs)

    def pull(self, *args, **kwargs):
        """
        Pull a container from the remote.
        """
        return self.remote.pull(*args, **kwargs)

    def login(
        self,
        username: str,
        password: str,
        password_stdin: bool = False,
        insecure: bool = False,
        hostname: Optional[str] = None,
        config_path: Optional[List[str]] = None,
    ):
        """
        Login to a registry.


        :param registry: if provided, use this custom provider instead of default
        :type registry: oras.provider.Registry or None

        :param username: the user account name
        :type username: str
        :param password: the user account password
        :type password: str
        :param password_stdin: get the password from standard input
        :type password_stdin: bool
        :param insecure: use http instead of https
        :type insecure: bool
        :param hostname: the hostname to login to
        :type hostname: str
        :param config_path: list of config paths to add
        :type config_path: list
        """
        login_func = main.login
        if hasattr(self.remote, "login"):
            login_func = self.remote.login  # type: ignore
        return login_func(
            username=username,
            password=password,
            password_stdin=password_stdin,
            insecure=insecure,
            hostname=hostname,
            config_path=config_path,  # type: ignore
        )

    def logout(self, hostname: str):
        """
        Logout from a registry, meaning removing any auth (if loaded)

        :param hostname: the hostname to login to
        :type hostname: str
        """
        self.remote.logout(hostname)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import sys
from typing import Optional

import oras.utils
from oras.logger import logger


def login(
    username: Optional[str] = None,
    password: Optional[str] = None,
    password_stdin: bool = False,
    insecure: bool = False,
    hostname: Optional[str] = None,
    config_path: Optional[str] = None,
):
    """
    Login to an OCI registry.

    The username and password can come from stdin.
    """
    client = oras.utils.get_docker_client(tls=not insecure)

    # Read password from stdin
    if password_stdin:
        password = readline()

    # No password provided
    elif not password:

        # No username, try to get from stdin
        if not username:
            username = input("Username: ")

        # if we still don't have a username, we require a token
        if not username:
            password = input("Token: ")
            if not password:
                logger.exit("token required")

        # If we do have a username, we just need a passowrd
        else:
            password = input("Password: ")
            if not password:
                logger.exit("password required")

    else:
        logger.warning(
            "WARNING! Using --password via the CLI is insecure. Use --password-stdin."
        )

    # Login
    # https://docker-py.readthedocs.io/en/stable/client.html?highlight=login#docker.client.DockerClient.login
    result = client.login(
        username=username,
        password=password,
        registry=hostname,
        dockercfg_path=config_path,
    )
    logger.info(result["Status"])


def readline() -> str:
    """
    Read lines from stdin
    """
    content = sys.stdin.readlines()
    return content[0].strip()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import base64
import os
import re
from typing import List, Optional

import oras.utils
from oras.logger import logger


def load_configs(configs: Optional[List[str]] = None):
    """
    Load one or more configs with credentials from the filesystem.

    :param configs: list of configuration paths to load, defaults to None
    :type configs: optional list
    """
    import docker

    configs = configs or []
    default_config = docker.context.config.find_config_file()

    # Add the default docker config
    if default_config:
        configs.append(default_config)
    configs = set(configs)  # type: ignore

    # Load configs until we find our registry hostname
    auths = {}
    for config in configs:
        if not os.path.exists(config):
            logger.warning(f"{config} does not exist.")
            continue
        cfg = oras.utils.read_json(config)
        auths.update(cfg.get("auths", {}))
    return auths


def get_basic_auth(username: str, password: str):
    """
    Prepare basic auth from a username and password.

    :param username: the user account name
    :type username: str
    :param password: the user account password
    :type password: str
    """
    auth_str = "%s:%s" % (username, password)
    return base64.b64encode(auth_str.encode("utf-8")).decode("utf-8")


class authHeader:
    def __init__(self, lookup: dict):
        """
        Given a dictionary of values, match them to class attributes

        :param lookup : dictionary of key,value pairs to parse into auth header
        :type lookup: dict
        """
        self.service: Optional[str] = None
        self.realm: Optional[str] = None
        self.scope: Optional[str] = None
        for key in lookup:
            if key in ["realm", "service", "scope"]:
                setattr(self, key, lookup[key])


def parse_auth_header(authHeaderRaw: str) -> authHeader:
    """
    Parse authentication header into pieces

    :param username: the user account name
    :type username: str
    :param password: the user account password
    :type password: str
    """
    regex = re.compile('([a-zA-z]+)="(.+?)"')
    matches = regex.findall(authHeaderRaw)
    lookup = dict()
    for match in matches:
        lookup[match[0]] = match[1]
    return authHeader(lookup)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import oras.client


def main(args, parser, extra, subparser):
    """
    A wrapper around an oras client push.
    """
    client = oras.client.OrasClient(insecure=args.insecure)
    client.push(
        config_path=args.config,
        disable_path_validation=args.disable_path_validation,
        files=args.filerefs,
        manifest_config=args.manifest_config,
        username=args.username,
        password=args.password,
        quiet=args.quiet,
        target=args.target,
    )



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"


def iter_localhosts(name: str):
    """
    Given a url with localhost, always resolve to 127.0.0.1.

    :param name : the name of the original host string
    :type name: str
    """
    names = [name]
    if "localhost" in name:
        names.append(name.replace("localhost", "127.0.0.1"))
    elif "127.0.0.1" in name:
        names.append(name.replace("127.0.0.1", "localhost"))
    for name in names:
        yield name


def get_docker_client(insecure: bool = False, **kwargs):
    """
    Get a docker client.

    :param tls : enable tls
    :type tls: bool
    """
    import docker

    return docker.DockerClient(tls=not insecure, **kwargs)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright The ORAS Authors."
__license__ = "Apache-2.0"

import oras.client


def main(args, parser, extra, subparser):
    """
    A wrapper around an oras client pull.
    """
    client = oras.client.OrasClient(insecure=args.insecure)
    client.pull(
        config_path=args.config,
        allowed_media_type=args.allowed_media_type
        if not args.allow_all_media_types
        else [],
        overwrite=not args.keep_old_files,
        manifest_config_ref=args.manifest_config_ref,
        outdir=args.output,
        password=args.password,
        username=args.username,
        target=args.target,
    )
